"""List all steam games and start them when selected"""
import os
import glob
import struct
import binascii
import logging
from lnxlink.modules.scripts.helpers import import_install_package, syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Steam"
        self.lnxlink = lnxlink
        self._requirements()
        self.steam_config = self._find_libary_config()
        if self.steam_config is None:
            raise SystemError("Steam not found")
        self.games = {}
        self.games = self._get_games()

    def _requirements(self):
        self.vdf = import_install_package("vdf", ">=3.4")

    def exposed_controls(self):
        """Exposes to home assistant"""
        mygames = list(self.games.values())
        mygames.sort()
        discovery_info = {
            "Game Select": {
                "type": "select",
                "icon": "mdi:steam",
                "options": mygames,
            },
        }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        games = self._get_games()
        if self.games != games:
            self.lnxlink.setup_discovery()
        self.games = games
        return self.games

    def start_control(self, topic, data):
        """Control system"""
        try:
            position = list(self.games.values()).index(data)
            game_id = list(self.games.keys())[position]
            logger.info("running command: steam steam://rungameid/%s", game_id)
            syscommand(f"steam steam://rungameid/{game_id}")
        except ValueError:
            logger.error("Can't find selected game: %s", data)

    def _find_libary_config(self, level=4):
        """Finds the library folders configuration from home directory
        with a max depth level"""
        home_dir = os.path.expanduser("~")
        num_sep = home_dir.count("/")
        name = "libraryfolders.vdf"
        for root, dirs, files in os.walk(home_dir):
            num_sep_this = root.count("/")
            if num_sep + level <= num_sep_this:
                del dirs[:]
            if name in files:
                return os.path.join(root, name)
        return None

    def _get_games(self):
        """Gets a dictionary of the game_ids and their names"""
        with open(self.steam_config, encoding="UTF-8") as file:
            libraries = self.vdf.load(file)

        # Steam Games
        games = {}
        for num in libraries["libraryfolders"].values():
            for app_id in num["apps"].keys():
                if app_id not in self.games:
                    acf_path = os.path.join(
                        num["path"], f"steamapps/appmanifest_{app_id}.acf"
                    )
                    with open(acf_path, encoding="UTF-8") as file:
                        game_properties = self.vdf.load(file)
                    games[app_id] = game_properties["AppState"]["name"]
                else:
                    games[app_id] = self.games[app_id]

        # Non-Steam Games
        shortcut_dirs = os.path.abspath(
            f"{self.steam_config}/../../userdata/*/config/shortcuts.vdf"
        )
        for shortcut_dir in glob.glob(shortcut_dirs):
            with open(shortcut_dir, "rb") as file:
                shortcuts_dict = self.vdf.binary_loads(file.read())
            for shortcut in shortcuts_dict["shortcuts"].values():
                # Convert to HEX
                bin_appid = struct.Struct("<i").pack(shortcut["appid"])
                hex_appid = binascii.hexlify(bin_appid).decode("UTF-8")

                # Convert to long_appid
                reversed_hex = bytes.fromhex(hex_appid)[::-1].hex()
                app_id = int(reversed_hex, 16) << 32 | 0x02000000
                games[app_id] = shortcut["AppName"]

        return games



"""Controls Docker instance"""
import logging
from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Docker"
        self.lnxlink = lnxlink
        self._requirements()
        try:
            # client = docker.from_env()
            self.client = self.docker.DockerClient(base_url="unix://run/docker.sock")
        except Exception as err:
            raise SystemError(f"Docker instance not found: {err}") from err
        self.containers = self._get_containers()

    def _requirements(self):
        self.docker = import_install_package("docker", ">=7.0.0")

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for container in self.containers:
            discovery_info[f"Docker {container}"] = {
                "type": "switch",
                "icon": "mdi:docker",
                "value_template": f"{{{{ value_json.get('{container}', {{}}).get('running') }}}}",
                "attributes_template": f"{{{{ value_json.get('{container}', {{}}) | tojson }}}}",
            }
        discovery_info["Docker Prune"] = {
            "type": "button",
            "icon": "mdi:docker",
        }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        containers = self._get_containers()
        if len(containers) != len(self.containers):
            self.lnxlink.setup_discovery()
        self.containers = containers
        return self.containers

    def _get_containers(self):
        include = self.lnxlink.config["settings"].get("docker", {}).get("include", [])
        exclude = self.lnxlink.config["settings"].get("docker", {}).get("exclude", [])
        containers = {}
        for container in self.client.containers.list(all=True):
            if len(include) > 0 and container.name not in include:
                continue
            if container.name in exclude:
                continue
            ports = set()
            for _, host in container.ports.items():
                if host is not None:
                    for host_info in host:
                        ports.add(host_info["HostPort"])
            running = "OFF"
            if container.attrs["State"]["Running"]:
                running = "ON"
            name_id = container.name.lower().replace(" ", "_")
            containers[name_id] = {
                "name": container.name,
                "image": ",".join(container.image.tags),
                "ports": list(ports),
                "running": running,
                "status": container.status,
            }
        return containers

    def start_control(self, topic, data):
        """Control system"""
        name_id = topic[1].replace("docker_", "")
        if name_id in self.containers:
            name = self.containers[name_id]["name"]
            if data == "ON":
                logger.info("Starting container %s", name)
                self.client.containers.get(name).start()
            elif data == "OFF":
                logger.info("Stopping container %s", name)
                self.client.containers.get(name).stop()
        elif name_id == "prune":
            # docker system prune -af
            logger.info("Running prune all")
            self.client.containers.prune()
            self.client.images.prune()
            self.client.networks.prune()
            self.client.volumes.prune()



"""Get Disk IO usage"""
import glob
import asyncio
from timeit import default_timer as timer


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "DiskIO"
        self.disks = self._get_disks()

        self.stat_items = [
            "read IOs",
            "read merges",
            "read sectors",
            "read ticks",
            "write IOs",
            "write merges",
            "write sectors",
            "write ticks",
            "in_flight",
            "io_ticks",
            "time_in_queue",
            "discard IOs",
            "discard merges",
            "discard sectors",
            "discard ticks",
            "flush IOs",
            "flush ticks",
        ]

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for disk in self.disks:
            discovery_info[f"Disk IO {disk}"] = {
                "type": "sensor",
                "icon": "mdi:barcode-scan",
                "unit": "%",
                "entity_category": "diagnostic",
                "state_class": "measurement",
                "value_template": f"{{{{ value_json.get('{disk}') }}}}",
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        disks = self._get_disks()
        if self.disks != disks:
            self.disks = disks
            self.lnxlink.setup_discovery("disk_io")
        results = asyncio.run(self._get_info_async())
        return results

    async def _get_info_async(self):
        features = []
        for disk in self.disks:
            features.append(self._run_check(disk))
        gathers = await asyncio.gather(*features)
        results = {}
        for disk, utilization in gathers:
            results[disk] = utilization
        return results

    def _get_disks(self):
        disks = []
        for disk in glob.glob("/sys/block/*", recursive=False):
            disk_name = disk.removeprefix("/sys/block/")
            excludes = ["loop", "ram", "zram"]
            if any(disk_name.startswith(x) for x in excludes):
                continue
            disks.append(disk_name)
        return disks

    async def _run_check(self, disk):
        with open(f"/sys/block/{disk}/stat", encoding="UTF-8") as file:
            pout1 = file.read()
        start = timer()
        await asyncio.sleep(0.1)
        with open(f"/sys/block/{disk}/stat", encoding="UTF-8") as file:
            pout2 = file.read()
        totaltime = timer() - start

        stats1 = dict(zip(self.stat_items, map(int, pout1.split())))
        stats2 = dict(zip(self.stat_items, map(int, pout2.split())))

        utilization = (stats2["io_ticks"] - stats1["io_ticks"]) / totaltime / 10
        utilization = min(utilization, 100)
        utilization = int(round(utilization, 0))
        return disk, utilization



"""Controls systemd services"""
import logging
from lnxlink.modules.scripts.helpers import syscommand, import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "SystemD"
        self.lnxlink = lnxlink
        self.services = self.lnxlink.config["settings"].get("systemd", [])
        self.services = [] if self.services is None else self.services
        if len(self.services) == 0:
            logger.info("No systemd settings found on configuration.")
        self._requirements()

    def _requirements(self):
        dasbus = import_install_package("dasbus", ">=1.7", "dasbus.connection")
        self.bus = dasbus.connection.SystemMessageBus()

    def get_info(self):
        """Gather information from the system"""
        info = {}
        for service in self.services:
            dbus_service = (
                service.replace("-", "_2d")
                .replace(".", "_2e")
                .replace("\\", "_5c")
                .replace("@", "_40")
            )
            proxy = self.bus.get_proxy(
                service_name="org.freedesktop.systemd1",
                object_path=f"/org/freedesktop/systemd1/unit/{dbus_service}",
                interface_name="org.freedesktop.systemd1.Unit",
            )
            status = "OFF"
            if proxy.ActiveState == "active":
                status = "ON"
            name = service.replace(".service", "")
            info[name] = status
        return info

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for service in self.services:
            name = service.replace(".service", "")
            discovery_info[f"Systemd {name}"] = {
                "type": "switch",
                "icon": "mdi:application-cog",
                "value_template": f"{{{{ value_json.get('{name}') }}}}",
            }
        return discovery_info

    def start_control(self, topic, data):
        """Control system"""
        service = topic[1].replace("systemd_", "") + ".service"
        dbus_service = (
            service.replace("-", "_2d")
            .replace(".", "_2e")
            .replace("\\", "_5c")
            .replace("@", "_40")
        )
        proxy = self.bus.get_proxy(
            service_name="org.freedesktop.systemd1",
            object_path=f"/org/freedesktop/systemd1/unit/{dbus_service}",
            interface_name="org.freedesktop.systemd1.Unit",
        )
        if data.lower() == "off":
            try:
                proxy.Stop("replace")
            except Exception:
                syscommand(f"sudo systemctl stop {service} &")
        elif data.lower() == "on":
            try:
                proxy.Start("replace")
            except Exception:
                syscommand(f"sudo systemctl start {service} &")
